--[[
Copyright (c) 2019 Raidho36 <coaxilthedrug@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
]]--

--[[
-- - - = = = = = = = = = = = = = = = - - --
-- - - = = = APPLICATION NOTES = = = - - --
-- - - = = = = = = = = = = = = = = = - - --

Unique values variant of gapless array. Enforcing is done through table lookup.
Values may be removed by value as well as by index.

Main intended operations are "push" and "pull"; gapless array is maintained,
which is beneficial for performance. Items may be randomly reordered during
pulling.

Normal behavior array manipulation is available via "insert" and "remove"
methods.
]]--

local Array = setmetatable ( { }, { __call = function ( class, ... ) return class.new ( ... ) end } )
Array.__index = Array

function Array.new ( )
	return setmetatable ( { len = 0, refs = { } }, Array )
end

function Array:finalize ( )
	self:clear ( )
	self.refs = nil
end

function Array:push ( val )
	assert ( type ( val ) ~= 'nil', "Value is nil." )

	local refs = self.refs

	if not refs[ val ] then
		local len = self.len + 1
		self.len = len
		self[ len ] = val
		refs[ val ] = len
	end
end

function Array:pop ( )
	local len = self.len
	if len == 0 then return nil end

	self.len = len - 1
	local refs = self.refs
	local obj = self[ len ]

	self[ len ] = nil
	refs[ obj ] = nil

	return obj
end

-- replaces selected element with the last element and trims the array
function Array:pull ( idx )
	assert ( type ( idx ) == 'number', "Element index is not a number." )
	assert ( idx >= 1 and idx <= self.len, "Element index out of range." )

	local refs = self.refs

	local len = self.len
	self.len = len - 1

	local idxobj = self[ idx ]
	local lenobj = self[ len ]

	self[ idx ], self[ len ] = lenobj, nil
	refs[ lenobj ], refs[ idxobj ] = idx, nil

	return idxobj
end
function Array:pullval ( idxobj )
	assert ( type ( idxobj ) ~= 'nil', "Value is nil." )

	local refs = self.refs
	local idx = refs[ idxobj ]
	if not idx then return nil end

	local len = self.len
	self.len = len - 1
	local lenobj = self[ len ]

	self[ idx ], self[ len ] = lenobj, nil
	refs[ lenobj ], refs[ idxobj ] = idx, nil

	return idxobj
end

-- insert an item in the middle, shifting everything past it one slot forward
function Array:insert ( idx, val )
	assert ( type ( val ) ~= 'nil', "Value is nil." )
	assert ( type ( idx ) == 'number', "Element index is not a number." )
	assert ( idx >= 1 and idx <= self.len, "Element index out of range." )

	local refs = self.refs
	local len = self.len + 1
	local vpos = refs[ val ]

	if vpos then -- value is already in the array
		if vpos > idx then --move array forwards
			for i = vpos, idx + 1, -1 do
				local v = self[ i - 1 ]
				refs[ v ], self[ i ] = i, v
			end
			self[ idx ], refs[ val ] = val, idx
		elseif vpos < idx then --move array backwards
			for i = vpos, idx - 1 do
				local v = self[ i + 1 ]
				refs[ v ], self[ i ] = i, v
			end
			self[ idx ], refs[ val ] = val, idx
		end -- do nothing if it's in its place
	else -- insert normally
		for i = len, idx + 1, -1 do
			local v = self[ i - 1 ]
			refs[ v ], self[ i ] = i, v
		end
		self[ idx ], refs[ val ], self.len = val, idx, len
	end
end

-- shifts the array one place to overwrite selected element and trims the array
function Array:remove ( idx )
	assert ( type ( idx ) == 'number', "Element index is not a number." )
	assert ( idx >= 1 and idx <= self.len, "Element index out of range." )

	local refs = self.refs
	local len = self.len
	local obj = self[ idx ]

	for i = idx, len - 1 do
		local v = self[ i + 1 ]
		refs[ v ], self[ i ] = i, v
	end
	refs[ obj ], self[ len ], self.len = nil, nil, len - 1

	return obj
end
function Array:removeval ( obj )
	assert ( type ( val ) ~= 'nil', "Value is nil." )

	local refs = self.refs
	local idx = refs[ obj ]
	if not idx then return nil end

	local len = self.len

	for i = idx, len - 1 do
		local v = self[ i + 1 ]
		refs[ v ], self[ i ] = i, v
	end
	refs[ obj ], self[ len ], self.len = nil, nil, len - 1

	return obj
end

function Array:findval ( val )
	assert ( type ( val ) ~= 'nil', "Value is nil." )

	return self.refs[ val ]
end

function Array:clear ( )
	local refs = self.refs

	for i = self.len, 1, -1 do
		local v = self[ i ]
		self[ i ], refs[ v ] = nil, nil
	end
	self.len = 0
end

return Array
